//---------------------------------------------------------------------------------
// File Name: DemoLotsOfBoxes.cpp
// Description: 	
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "../Interface/IApp.h"
#include "../Interface/IDemoPlugin.h"
#include "../Interface/IMouse.h"
#include "../Interface/ICamera.h"

#include "../TA/Common/RandFunc.h"

#include "../TA/Physics/Physics.h"
#include "../TA/Physics/CollisionObjectAABBMesh.h"
#include "../TA/Physics/CollisionObjectCombo.h"

#include "../Shared/Landscape.h"
#include "../Shared/DemoPhysicsRender.h"

#include <windows.h>

using namespace TA;
	
static bool KeyIsDown(int nKey) { return (GetAsyncKeyState(nKey) & 0x8000) != 0; }
static bool KeyIsJustPressed(int nKey) { return (GetAsyncKeyState(nKey) & 0x8001) == 0x8001; }

class DemoLotsOfBoxes : public IDemoPlugin
{
public:
	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	DemoLotsOfBoxes()
	{
		m_pApp = 0;
		m_pMouse = 0;
		m_pCamera = 0;
		m_pLandscape = 0;

		m_nMaxNumDynamicObjects = 0;
		m_ppDynamicObjectList = 0;

		m_fTimeUntilNextBoxAppear = 0.0f;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	~DemoLotsOfBoxes()
	{
		Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Release() { delete this; }

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Initialise(IApp* pApp)
	{
		m_pApp = pApp;

		// Implement physics debug rendering functions.
		DemoPhysicsRender::Initialise(pApp);

		// Initialise mouse.
		m_pMouse = m_pApp->CreateMouse();
		m_pMouse->Initialise(true);

		// Initialise camera.
		m_pCamera = m_pApp->CreateCamera();
		m_pCamera->Initialise();

		// Initialise Physics.
		AABB worldAABB;
		worldAABB.Initialise(k_v3Zero, Vec3(5000.0f, 5000.0f, 5000.0f));
		Physics::CreateInstance(worldAABB);	

		//---------------------------------------------------------------------------------
		// Initialise Landscape
		//---------------------------------------------------------------------------------

		// Initialise Landscape collision object.
		m_pLandscape = new Landscape();
		m_pLandscape->Initialise(m_pApp, 0.0f);

		// Create a static object for the collision.
		StaticObject* pStaticObject = StaticObject::CreateNew();
		pStaticObject->Initialise(m_pLandscape->GetCollisionObject());

		// Add the static object to Physics.
		Physics::GetInstance().AddStaticObject(pStaticObject);
		// If we wanted to keep call Physics::GetInstance().RemoveStaticObject(pStaticObject) 
		// later, we would have to keep a pointer to pStaticObject.
		pStaticObject->Release(); 
		pStaticObject = 0;



		//---------------------------------------------------------------------------------
		// Initialise Dynamic Objects
		//---------------------------------------------------------------------------------

		// Initialise collision object
		CollisionObjectCombo* pCollisionObjectCombo = CollisionObjectCombo::CreateNew();
		pCollisionObjectCombo->InitialiseAsABox(1.0f, 1.0f, 1.0f);

		Vec3 v3GroundPos(1.0f, 0.0f, 1.0f);
		// Get placement height
		Collision collision = Physics::GetInstance().TestLineForCollision(
			v3GroundPos + k_v3UnitY * 1000.0f,
			-k_v3UnitY * 2000.0f,
			Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);

		if (collision.CollisionOccurred())
			v3GroundPos = collision.GetPosition();
		v3GroundPos.y += 2.0f;
			
		
		// Initialise objects
		m_nNumDynamicObjects = 0;
		m_nMaxNumDynamicObjects = 201;
		m_ppDynamicObjectList = new DynamicObject* [m_nMaxNumDynamicObjects];
		for (int nDynamicObject = 0; nDynamicObject < m_nMaxNumDynamicObjects; nDynamicObject++)
		{
			m_ppDynamicObjectList[nDynamicObject] = DynamicObject::CreateNew();
			TA_ASSERT(m_ppDynamicObjectList[nDynamicObject]);
			DynamicObject& dynamicObject = *m_ppDynamicObjectList[nDynamicObject];
			dynamicObject.Initialise(pCollisionObjectCombo);
			dynamicObject.SetMass(10.0f);
		}
		pCollisionObjectCombo->Release();

		{
			TA_ASSERT(m_ppDynamicObjectList[m_nNumDynamicObjects]);
			DynamicObject& dynamicObject = *m_ppDynamicObjectList[m_nNumDynamicObjects];

			Vec3 v3Pos = v3GroundPos;
			v3Pos.x += (float)(nDynamicObject & 1) * 0.01f;
			v3Pos.z += (float)(nDynamicObject & 1) * 0.01f;
			dynamicObject.SetFrame(
				MFrame(
					v3Pos,
					Mat33(Mat33::ROTATION_AXIS_Y, k_fPi * 0.0f * nDynamicObject)));
			v3GroundPos.y += 3.0f;
			Physics::GetInstance().AddDynamicObject(&dynamicObject);	

			m_nNumDynamicObjects++;

			m_fTimeUntilNextBoxAppear = 1.0f;

		}

		m_bHelp = true;
		m_bInvertMouse = true;
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Finalise()
	{
		if (m_ppDynamicObjectList)
		{
			for (int i = 0; i < m_nMaxNumDynamicObjects; i++)
			{
				if (m_ppDynamicObjectList[i])
				{
					m_ppDynamicObjectList[i]->Release();
					m_ppDynamicObjectList[i] = 0;
				}
			}
			delete []m_ppDynamicObjectList;
			m_ppDynamicObjectList = 0;
		}

		if (m_pLandscape)
		{
			delete m_pLandscape;
			m_pLandscape = 0;
		}
		if (m_pMouse)
		{
			m_pMouse->Release();
			m_pMouse = 0;
		}
		if (m_pCamera)
		{
			m_pCamera->Release();
			m_pCamera = 0;
		}
		m_pApp = 0;

		Physics::DestroyInstance();

		DemoPhysicsRender::Finalise();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Update(float fDt)
	{
		TA_ASSERT(m_ppDynamicObjectList);

		// Update Input
		m_pMouse->Update();
		if (m_pMouse->IsLeftButtonDown())
		{		
			TA_ASSERT(m_ppDynamicObjectList[0]);
			Physics::GetInstance().SetDynamicObjectIsMoving(m_ppDynamicObjectList[0]);
			m_ppDynamicObjectList[0]->ApplyImpulse(
				m_pCamera->GetRotation().v3Z * 7.5f,			// Impulse.
				m_ppDynamicObjectList[0]->GetPosition());	// World position.
		}

		if (KeyIsJustPressed(VK_F1))
			m_bHelp = !m_bHelp;
		if (KeyIsJustPressed('I'))
			m_bInvertMouse = !m_bInvertMouse;
		
		m_fTimeUntilNextBoxAppear -= fDt;
		if (m_fTimeUntilNextBoxAppear < 0.0f)
		{
			m_fTimeUntilNextBoxAppear = 2.0f;

			const float k_fRandRange = 100.0f;
			Vec3 v3GroundPos(
				RandFunc::RandBetweenPosNeg(k_fRandRange), 
				0.0f, 
				RandFunc::RandBetweenPosNeg(k_fRandRange));
			// Get placement height
			Collision collision = Physics::GetInstance().TestLineForCollision(
				v3GroundPos + k_v3UnitY * 1000.0f,
				-k_v3UnitY * 2000.0f,
				Physics::FLAG_ALL_NON_DYNAMIC_OBJECTS);

			if (collision.CollisionOccurred())
				v3GroundPos = collision.GetPosition();
			v3GroundPos.y += 2.0f;

			for (int i = 0; i < 4; i++)
			{
				if (m_nNumDynamicObjects >= m_nMaxNumDynamicObjects)
					break;

				TA_ASSERT(m_ppDynamicObjectList[m_nNumDynamicObjects]);
				DynamicObject& dynamicObject = *m_ppDynamicObjectList[m_nNumDynamicObjects];

				Vec3 v3Pos = v3GroundPos;
				v3Pos.x += (float)(i & 1) * 0.01f;
				v3Pos.z += (float)(i & 1) * 0.01f;
				dynamicObject.SetFrame(
					MFrame(
						v3Pos,
						Mat33(Mat33::ROTATION_AXIS_Y, k_fPi * 0.0f * i)));
				v3GroundPos.y += 0.98f;
				Physics::GetInstance().AddDynamicObject(&dynamicObject);	

				m_nNumDynamicObjects++;

				m_fTimeUntilNextBoxAppear = 0.25f;
			}
		}

		// Update physics
		Physics::GetInstance().Update(fDt);

		// Update the camera.	
		m_pCamera->SetNearClip(0.1f);
		m_pCamera->SetFarClip(1000.0f);
		//m_pCamera->SetFOV(
		m_pCamera->SetViewDistance(5.0f);
		m_pCamera->UpdateRotation(
			m_pMouse->GetXMovement(),
			m_bInvertMouse ? m_pMouse->GetYMovement() : -m_pMouse->GetYMovement());		
		TA_ASSERT(m_ppDynamicObjectList[0]);
		m_pCamera->SetTargetPosition(
			m_ppDynamicObjectList[0]->GetPosition() + k_v3UnitY);
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void Render()
	{
		// Set the camera
		m_pCamera->Realise();
		
		// Render physics (renders collision objects)
		Physics::GetInstance().Render();

		// Render the Landscape
		m_pLandscape->Render();

		RenderHelp();
	}

	//---------------------------------------------------------------------------------
	//---------------------------------------------------------------------------------
	void RenderHelp()
	{
		if (m_bHelp)
		{
			m_pApp->RenderText("Toggle help on and off (F1)");
			m_pApp->RenderText("Move the mouse to rotate the camera");
			m_pApp->RenderText("Invert mouse (I)");
			m_pApp->RenderText("Move object (Left Mouse Button)");
			m_pApp->RenderText("Quit (Esc)");
		}
	}

private:
	IApp* m_pApp;
	IMouse* m_pMouse;	
	ICamera* m_pCamera;
	Landscape* m_pLandscape;

	int m_nMaxNumDynamicObjects;
	int m_nNumDynamicObjects;
	DynamicObject** m_ppDynamicObjectList;
	float m_fTimeUntilNextBoxAppear;
	bool m_bHelp;
	bool m_bInvertMouse;
};

TA_IMPLEMENT_DEMO_PLUGIN(DemoLotsOfBoxes);
