//---------------------------------------------------------------------------------
// File Name: RagDoll.cpp
// Description:
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#include "RagDoll.h"

#include "RagDoll.h"

#include "../TA/Physics/Physics.h"
#include "../TA/Physics/CollisionObjectCombo.h"
#include "../TA/Physics/CollisionObjectCapsule.h"

#include <memory.h>

using namespace TA;

#include <memory.h>

using namespace TA;

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
RagDoll::RagDoll()
{
	for (int i = 0; i < NUM_SEGMENTS; i++)	
	{
		m_ppObjectList[i] = 0;
		m_pFrameList[i].Clear();
	}	
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
RagDoll::~RagDoll()
{
	Finalise();
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::Initialise()
{	
	const float k_fRestTimeMultiplier = 2.0f;
	const bool k_bExtraStability = true;
	float fTargetMass = 80.0f;

	const float k_fScale = 1.0f;
	const float k_fShoulderWidth = 0.2f;
	Vec3 v3PosHead = Vec3(0.0f, 1.6f, 0.0f) * k_fScale;
	Vec3 v3PosULArm = Vec3(-k_fShoulderWidth, 1.55f, 0.0f) * k_fScale;
	Vec3 v3PosURArm = Vec3(k_fShoulderWidth, 1.55f, 0.0f) * k_fScale;
	Vec3 v3PosMLArm = Vec3(-k_fShoulderWidth, 1.2f, 0.0f) * k_fScale;
	Vec3 v3PosMRArm = Vec3(k_fShoulderWidth, 1.2f, 0.0f) * k_fScale;
	Vec3 v3PosLLArm = Vec3(-k_fShoulderWidth, 0.8f, 0.0f) * k_fScale;
	Vec3 v3PosLRArm = Vec3(k_fShoulderWidth, 0.8f, 0.0f) * k_fScale;
	//Vec3 v3PosLElbowPtr = Vec3() * k_fScale;
	//Vec3 v3PosRElbowPtr = Vec3() * k_fScale;

	
	const float k_fLegWidth = 0.1f;
	Vec3 v3PosULLeg = Vec3(-k_fLegWidth, 1.0f, 0.0f) * k_fScale;
	Vec3 v3PosURLeg = Vec3(k_fLegWidth, 1.0f, 0.0f) * k_fScale;
	Vec3 v3PosMLLeg = Vec3(-k_fLegWidth, 0.5f, 0.0f) * k_fScale;
	Vec3 v3PosMRLeg = Vec3(k_fLegWidth, 0.5f, 0.0f) * k_fScale;
	Vec3 v3PosLLLeg = Vec3(-k_fLegWidth, 0.0f, 0.0f) * k_fScale;
	Vec3 v3PosLRLeg = Vec3(k_fLegWidth, 0.0f, 0.0f) * k_fScale;

	Vec3 v3PosMiddle = (v3PosULArm + v3PosURArm + v3PosULLeg + v3PosURLeg) * 0.25f;
	
	MFrame mFrame;

	const float k_fLengthMod = 1.0f;
	const float k_fRadiusMod = 0.9f;
	const float k_fIKShortenMod = 0.95f;
	const float k_fFriction = 1.0f;
	const bool k_bNoRigidFriction = true;

	enum 
	{
		PART_BODY = 0,
		PART_LOWER,
		PART_HEAD,
		PART_UL_ARM,
		PART_UR_ARM,
		PART_LL_ARM,
		PART_LR_ARM,
		PART_UL_LEG,
		PART_UR_LEG,
		PART_LL_LEG,
		PART_LR_LEG,
	};

	
	int nUserGroupID = DynamicObject::CreateUserGroup();
	//---------------------------------------------------------------------------------	
	// Body
	//---------------------------------------------------------------------------------
	DynamicObject* pBody = 0;
	{
		Vec3 v3Top = 
			(v3PosULArm +
			v3PosURArm) * 0.5f;
		Vec3 v3Bottom = 
			(v3PosULLeg +
			v3PosURLeg) * 0.5f;

		TA_ASSERT(v3Top.y > v3Bottom.y);
		float fHeight = (v3Top.y > v3Bottom.y);
		fHeight *= 0.5f;
		pBody = DynamicObject::CreateNew();
		DynamicObject& object = *pBody;
		object.InitialiseAsACapsule(
			Vec3(0.0f, fHeight * k_fLengthMod * 0.25f, 0.0f),
			Vec3(0.0f, -fHeight * k_fLengthMod * 0.20f, 0.0f),
			0.17f * k_fRadiusMod);
		mFrame.v3Translation =  ((v3Top + v3Bottom) * 0.5f + v3Top) * 0.5f;
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_BODY);		
		object.DisallowCollisionWithUserGroupItemId(PART_LOWER);
		object.DisallowCollisionWithUserGroupItemId(PART_HEAD);
		object.DisallowCollisionWithUserGroupItemId(PART_UL_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UR_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UL_LEG);
		object.DisallowCollisionWithUserGroupItemId(PART_UR_LEG);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);
	}

	//---------------------------------------------------------------------------------	
	// Body Lower
	//---------------------------------------------------------------------------------
	DynamicObject* pBodyLower = 0;
	{
		Vec3 v3Top = 
			(v3PosULArm +
			v3PosURArm) * 0.5f;
		Vec3 v3Bottom = 
			(v3PosULLeg +
			v3PosURLeg) * 0.5f;

		TA_ASSERT(v3Top.y > v3Bottom.y);
		float fHeight = (v3Top.y > v3Bottom.y);
		fHeight *= 0.5f;
		pBodyLower = DynamicObject::CreateNew();
		DynamicObject& object = *pBodyLower;
		object.InitialiseAsACapsule(
			Vec3(0.0f, fHeight * k_fLengthMod * 0.20f, 0.0f),
			Vec3(0.0f, -fHeight * k_fLengthMod * 0.25f, 0.0f),
			0.15f * k_fRadiusMod);
		mFrame.v3Translation = ((v3Top + v3Bottom) * 0.5f + v3Bottom) * 0.5f;
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_LOWER);		
		object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_HEAD);
		object.DisallowCollisionWithUserGroupItemId(PART_UL_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UR_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UL_LEG);
		object.DisallowCollisionWithUserGroupItemId(PART_UR_LEG);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);
	}
	
	//---------------------------------------------------------------------------------	
	// Head
	//---------------------------------------------------------------------------------
	DynamicObject* pHead = 0;
	{
		Vec3 v3Bottom = v3PosHead;
		float fHeight = 0.18f;
		pHead = DynamicObject::CreateNew();
		DynamicObject& object = *pHead;
		object.InitialiseAsACapsule(
			Vec3(0.0f, fHeight * k_fLengthMod * 0.5f, 0.0f),
			Vec3(0.0f, -fHeight * k_fLengthMod * 0.5f, 0.0f),
			0.1f * k_fRadiusMod);
		mFrame.v3Translation.Initialise(v3Bottom.x, v3Bottom.y + fHeight, v3Bottom.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_HEAD);
		object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_LOWER);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);
	}
		
	//---------------------------------------------------------------------------------	
	// Upper Left Arm
	//---------------------------------------------------------------------------------
	DynamicObject* pULArm = 0;
	Vec3 v3LeftArmTop;
	Vec3 v3LeftArmBottom;
	{
		const Vec3& v3Top = v3PosULArm;
		const Vec3& v3Bottom = v3PosMLArm;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		pULArm = DynamicObject::CreateNew();
		DynamicObject& object = *pULArm;
		object.InitialiseAsACapsule(
			Vec3(fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			Vec3(-fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			0.09f * k_fRadiusMod);
		mFrame.v3Translation.Initialise(v3Top.x - fLength * 0.5f, v3Top.y, v3Top.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_UL_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_LOWER);
		object.DisallowCollisionWithUserGroupItemId(PART_LL_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UR_ARM);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);

		v3LeftArmTop = v3Top;
		v3LeftArmBottom = v3Top;
		v3LeftArmBottom.x -= fLength;
	}
		
	//---------------------------------------------------------------------------------	
	// Upper Right Arm
	//---------------------------------------------------------------------------------
	DynamicObject* pURArm = 0;
	Vec3 v3RightArmTop;
	Vec3 v3RightArmBottom;
	{
		const Vec3& v3Top = v3PosURArm;
		const Vec3& v3Bottom = v3PosMRArm;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		pURArm = DynamicObject::CreateNew();
		DynamicObject& object = *pURArm;
		object.InitialiseAsACapsule(
			Vec3(fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			Vec3(-fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			0.09f * k_fRadiusMod);
		mFrame.v3Translation.Initialise(v3Top.x + fLength * 0.5f, v3Top.y, v3Top.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_UR_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_LOWER);
		object.DisallowCollisionWithUserGroupItemId(PART_LR_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UL_ARM);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);

		v3RightArmTop = v3Top;
		v3RightArmBottom = v3Top;
		v3RightArmBottom.x += fLength;
	}
	
	//---------------------------------------------------------------------------------	
	// Lower Left Arm
	//---------------------------------------------------------------------------------
	DynamicObject* pLLArm = 0;
	{
		const Vec3& v3Top = v3PosMLArm;
		const Vec3& v3Bottom = v3PosLLArm;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		pLLArm = DynamicObject::CreateNew();
		DynamicObject& object = *pLLArm;
		object.InitialiseAsACapsule(
			Vec3(fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			Vec3(-fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			0.08f * k_fRadiusMod);
		mFrame.v3Translation.Initialise(
			v3LeftArmBottom.x - fLength * 0.5f, 
			v3LeftArmBottom.y, 
			v3LeftArmBottom.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_LL_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UL_ARM);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);
	}
		
	//---------------------------------------------------------------------------------	
	// Lower Right Arm
	//---------------------------------------------------------------------------------
	DynamicObject* pLRArm = 0;
	{
		const Vec3& v3Top = v3PosMRArm;
		const Vec3& v3Bottom = v3PosLRArm;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		pLRArm = DynamicObject::CreateNew();
		DynamicObject& object = *pLRArm;
		object.InitialiseAsACapsule(
			Vec3(fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			Vec3(-fLength * k_fLengthMod * 0.5f, 0.0f, 0.0f),
			0.08f * k_fRadiusMod);
		mFrame.v3Translation.Initialise(
			v3RightArmBottom.x + fLength * 0.5f, 
			v3RightArmBottom.y, 
			v3RightArmBottom.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_LR_ARM);
		object.DisallowCollisionWithUserGroupItemId(PART_UR_ARM);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);
	}
	
		
	//---------------------------------------------------------------------------------	
	// Upper Left Leg
	//---------------------------------------------------------------------------------
	DynamicObject* pULLeg = 0;
	Vec3 v3LeftLegTop;
	Vec3 v3LeftLegBottom;
	{
		const Vec3& v3Top = v3PosULLeg;
		const Vec3& v3Bottom = v3PosMLLeg;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		pULLeg = DynamicObject::CreateNew();
		DynamicObject& object = *pULLeg;
		object.InitialiseAsACapsule(
			Vec3(0.0f, fLength * k_fLengthMod * 0.5f, 0.0f),
			Vec3(0.0f, -fLength * k_fLengthMod * 0.5f, 0.0f),
			0.1f * k_fRadiusMod);
		mFrame.v3Translation.Initialise(v3Top.x, v3Top.y - fLength * 0.5f, v3Top.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_UL_LEG);
		object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_LOWER);
		object.DisallowCollisionWithUserGroupItemId(PART_LL_LEG);
		//object.DisallowCollisionWithUserGroupItemId(PART_UR_LEG);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);

		v3LeftLegTop = v3Top;
		v3LeftLegBottom = v3Top;
		v3LeftLegBottom.y -= fLength;
	}
		
	//---------------------------------------------------------------------------------	
	// Upper Right Leg
	//---------------------------------------------------------------------------------
	DynamicObject* pURLeg = 0;
	Vec3 v3RightLegTop;
	Vec3 v3RightLegBottom;
	{
		const Vec3& v3Top = v3PosURLeg;
		const Vec3& v3Bottom = v3PosMRLeg;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		pURLeg = DynamicObject::CreateNew();
		DynamicObject& object = *pURLeg;
		object.InitialiseAsACapsule(
			Vec3(0.0f, fLength * k_fLengthMod * 0.5f, 0.0f),
			Vec3(0.0f, -fLength * k_fLengthMod * 0.5f, 0.0f),
			0.1f * k_fRadiusMod);
		mFrame.v3Translation.Initialise(v3Top.x, v3Top.y - fLength * 0.5f, v3Top.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_UR_LEG);
		object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_LOWER);
		object.DisallowCollisionWithUserGroupItemId(PART_LR_LEG);
		//object.DisallowCollisionWithUserGroupItemId(PART_UL_LEG);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);

		v3RightLegTop = v3Top;
		v3RightLegBottom = v3Top;
		v3RightLegBottom.y -= fLength;
	}
	
	//---------------------------------------------------------------------------------	
	// Lower Left Leg
	//---------------------------------------------------------------------------------
	DynamicObject* pLLLeg = 0;
	{
		const Vec3& v3Top = v3PosMLLeg;
		const Vec3& v3Bottom = v3PosLLLeg;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		pLLLeg = DynamicObject::CreateNew();
		DynamicObject& object = *pLLLeg;

		CollisionObjectCombo* pCollisionObjectCombo = CollisionObjectCombo::CreateNew();
		pCollisionObjectCombo->InitialiseAsACapsule(
			Vec3(0.0f, fLength * k_fLengthMod * 0.5f, 0.0f),
			Vec3(0.0f, -fLength * k_fLengthMod * 0.45f, 0.0f),
			0.08f * k_fRadiusMod);
		// Add a foot
		CollisionObjectCapsule* pCollisionA = CollisionObjectCapsule::CreateNew();
		pCollisionA->Initialise(
			Vec3(0.0f, -fLength * k_fLengthMod * 0.5f, 0.0f),
			Vec3(0.0f, -fLength * k_fLengthMod * 0.54f, 0.15f * k_fLengthMod),
			0.07f * k_fRadiusMod);
		pCollisionObjectCombo->AddCollisionObject(pCollisionA);
		object.Initialise(pCollisionObjectCombo);
		pCollisionA->Release();
		pCollisionObjectCombo->Release();


		mFrame.v3Translation.Initialise(
			v3LeftLegBottom.x, 
			v3LeftLegBottom.y - fLength * 0.5f, 
			v3LeftLegBottom.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_LL_LEG);
	//	object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_UL_LEG);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);

		
	}
		
	//---------------------------------------------------------------------------------	
	// Lower Right Leg
	//---------------------------------------------------------------------------------
	DynamicObject* pLRLeg = 0;
	{
		const Vec3& v3Top = v3PosMRLeg;
		const Vec3& v3Bottom = v3PosLRLeg;
		float fLength = (v3Top - v3Bottom).GetMagnitude();
		TA_ASSERT(pLRLeg == 0);
		pLRLeg = DynamicObject::CreateNew();
		DynamicObject& object = *pLRLeg;

		CollisionObjectCombo* pCollisionObjectCombo = CollisionObjectCombo::CreateNew();
		pCollisionObjectCombo->InitialiseAsACapsule(
			Vec3(0.0f, fLength * k_fLengthMod * 0.5f, 0.0f),
			Vec3(0.0f, -fLength * k_fLengthMod * 0.45f, 0.0f),
			0.08f * k_fRadiusMod);
		// Add a foot
		CollisionObjectCapsule* pCollisionA = CollisionObjectCapsule::CreateNew();
		pCollisionA->Initialise(
			Vec3(0.0f, -fLength * k_fLengthMod * 0.5f, 0.0f),
			Vec3(0.0f, -fLength * k_fLengthMod * 0.54f, 0.15f * k_fLengthMod),
			0.07f * k_fRadiusMod);
		pCollisionObjectCombo->AddCollisionObject(pCollisionA);
		object.Initialise(pCollisionObjectCombo);
		pCollisionA->Release();
		pCollisionObjectCombo->Release();

		mFrame.v3Translation.Initialise(
			v3RightLegBottom.x, 
			v3RightLegBottom.y - fLength * 0.5f, 
			v3RightLegBottom.z);
		mFrame.m33Rotation.SetToIdentity();
		object.SetFrame(mFrame);
		object.SetRigidFrictionDisabled(k_bNoRigidFriction);
		object.SetUserGroup(nUserGroupID);
		object.SetUserGroupItemId(PART_LR_LEG);
	//	object.DisallowCollisionWithUserGroupItemId(PART_BODY);
		object.DisallowCollisionWithUserGroupItemId(PART_UR_LEG);
		object.SetRestTimeMultiplier(k_fRestTimeMultiplier);
		object.SetExtraStability(k_bExtraStability);
		object.SetFriction(k_fFriction);

	}

	PhysicsJoint* pJoint = 0;
		
	Mat33 jointOrientation;
	Vec3 v3JointPos;
	//---------------------------------------------------------------------------------	
	// Neck Joint
	//---------------------------------------------------------------------------------	
	jointOrientation.SetToLookAt(k_v3UnitZ, k_v3UnitY);
	v3JointPos = v3PosHead;
	pJoint = &pBody->AddJoint(pHead);	
	pJoint->InitialiseLimitedBallAndSocket(
		v3JointPos,
		jointOrientation,
		k_fPi * 0.1f,
		k_fPi * 0.2f);
	
	//---------------------------------------------------------------------------------	
	// Middle
	//---------------------------------------------------------------------------------	
	jointOrientation.SetToLookAt(k_v3UnitZ, -k_v3UnitY);
	v3JointPos = v3PosMiddle;
	pJoint = &pBody->AddJoint(pBodyLower);	
	pJoint->InitialiseSquareSocket(
		v3JointPos,
		jointOrientation,
		-k_fPi * 0.1f,
		k_fPi * 0.3f,
		-k_fPi * 0.25f,
		k_fPi * 0.25f,
		-k_fPi * 0.2f,
		k_fPi * 0.2f);

	//---------------------------------------------------------------------------------	
	// Arm Joints
	//---------------------------------------------------------------------------------		
	jointOrientation.SetToLookAt(k_v3UnitY, -k_v3UnitX);
	v3JointPos = v3LeftArmTop;
	pJoint = &pBody->AddJoint(pULArm);
	pJoint->InitialiseSquareSocket(
		v3JointPos,
		jointOrientation,
		-k_fPi * 0.5f,
		k_fPi * 0.5f,
		-k_fPi * 0.25f,
		k_fPi * 0.25f,
		-k_fPi * 0.6f,
		k_fPi * 0.15f);

	jointOrientation.SetToLookAt(k_v3UnitY, k_v3UnitX);
	v3JointPos = v3RightArmTop;
	pJoint = &pBody->AddJoint(pURArm);
	pJoint->InitialiseSquareSocket(
		v3JointPos,
		jointOrientation,
		-k_fPi * 0.5f,
		k_fPi * 0.5f,
		-k_fPi * 0.25f,
		k_fPi * 0.25f,
		-k_fPi * 0.15f,
		k_fPi * 0.6f);

	jointOrientation.SetToLookAt(-k_v3UnitX, k_v3UnitY);
	v3JointPos = v3LeftArmBottom;
	pJoint = &pULArm->AddJoint(pLLArm);
	pJoint->InitialiseEulerConstraint(
		v3JointPos,
		jointOrientation,
		EulerAngles(0.0f, -k_fPi * 0.85f, 0.0f),
		EulerAngles(0.0f, 0.0f, 0.0f));

	jointOrientation.SetToLookAt(k_v3UnitX, k_v3UnitY);
	v3JointPos = v3RightArmBottom;
	pJoint = &pURArm->AddJoint(pLRArm);
	pJoint->InitialiseEulerConstraint(
		v3JointPos,
		jointOrientation,
		EulerAngles(0.0f, 0.0f, 0.0f),
		EulerAngles(0.0f, k_fPi * 0.85f, 0.0f));

	
	//---------------------------------------------------------------------------------	
	// Leg Joints
	//---------------------------------------------------------------------------------		
	
	jointOrientation.SetToLookAt(k_v3UnitZ, -k_v3UnitY);
	v3JointPos = v3LeftLegTop;
	pJoint = &pBodyLower->AddJoint(pULLeg);
	pJoint->InitialiseSquareSocket(
		v3JointPos,
		jointOrientation,
		-k_fPi * 0.05f,
		k_fPi * 0.55f,
		-k_fPi * 0.25f,
		k_fPi * 0.05f,
		-k_fPi * 0.1f,
		k_fPi * 0.2f);

	jointOrientation.SetToLookAt(k_v3UnitZ, -k_v3UnitY);
	v3JointPos = v3RightLegTop;
	pJoint = &pBodyLower->AddJoint(pURLeg);
	pJoint->InitialiseSquareSocket(
		v3JointPos,
		jointOrientation,
		-k_fPi * 0.05f,
		k_fPi * 0.55f,
		-k_fPi * 0.05f,
		k_fPi * 0.25f,
		-k_fPi * 0.2f,
		k_fPi * 0.1f);

	jointOrientation.SetToLookAt(-k_v3UnitY, -k_v3UnitX);
	v3JointPos = v3LeftLegBottom;
	pJoint = &pULLeg->AddJoint(pLLLeg);
	pJoint->InitialiseEulerConstraint(
		v3JointPos,
		jointOrientation,
		EulerAngles(0.0f, 0.0f, 0.0f),
		EulerAngles(0.0f, k_fPi * 0.85f, 0.0f));

	jointOrientation.SetToLookAt(-k_v3UnitY, -k_v3UnitX);
	v3JointPos = v3RightLegBottom;
	pJoint = &pURLeg->AddJoint(pLRLeg);
	pJoint->InitialiseEulerConstraint(
		v3JointPos,
		jointOrientation,
		EulerAngles(0.0f, 0.0f, 0.0f),
		EulerAngles(0.0f, k_fPi * 0.85f, 0.0f));

	
	m_ppObjectList[SEGMENT_HEAD] = pHead;
	m_ppObjectList[SEGMENT_BODY] = pBody;
	m_ppObjectList[SEGMENT_BODY_LOWER] = pBodyLower;
	m_ppObjectList[SEGMENT_ARM_UPPER_LEFT] = pULArm;
	m_ppObjectList[SEGMENT_ARM_UPPER_RIGHT] = pURArm;
	m_ppObjectList[SEGMENT_ARM_LOWER_LEFT] = pLLArm;
	m_ppObjectList[SEGMENT_ARM_LOWER_RIGHT] = pLRArm;
	m_ppObjectList[SEGMENT_LEG_UPPER_LEFT] = pULLeg;
	m_ppObjectList[SEGMENT_LEG_UPPER_RIGHT] = pURLeg;
	m_ppObjectList[SEGMENT_LEG_LOWER_LEFT] = pLLLeg;
	m_ppObjectList[SEGMENT_LEG_LOWER_RIGHT] = pLRLeg;

	float fTotalMass = 0.0f;

	for (int i = 0; i < NUM_SEGMENTS; i++)
	{
		fTotalMass += m_ppObjectList[i]->GetMass();
		m_pFrameList[i] = m_ppObjectList[i]->GetFrame();
	}

	float fMassMult = fTargetMass / fTotalMass;
	for (int i = 0; i < NUM_SEGMENTS; i++)
		 m_ppObjectList[i]->SetMass(fMassMult * m_ppObjectList[i]->GetMass());
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::Finalise()
{	
	for (int i = 0; i < NUM_SEGMENTS; i++)
	{
		DestroyDynamicObject(m_ppObjectList[i]);
		m_pFrameList[i].Clear();
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::Render()
{
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::Update(float fDt)
{	
	// Do a bit of velocity damping.
	int nObject;
	Vec3 v3TotalLinearMomentum;
	v3TotalLinearMomentum.Clear();
	float fTotalMass = 0.0f;
	for (nObject = 0; nObject < NUM_SEGMENTS; nObject++)
	{
		DynamicObject& object = *m_ppObjectList[nObject];
		fTotalMass += object.GetMass();
		v3TotalLinearMomentum += object.GetLinearVelocity() * object.GetMass();
	}
	TA_ASSERT(fTotalMass > 0.0f);
	Vec3 v3AverageLinearVelocity = v3TotalLinearMomentum / fTotalMass;	
	float fRateOfChange = (1.0f - Exp(-fDt * 10.0f));
	const float k_fFriction = 0.08f;
	const float k_fAngularFriction = 0.05f;
	for (nObject = 0; nObject < NUM_SEGMENTS; nObject++)
	{
		DynamicObject& object = *m_ppObjectList[nObject];

		Vec3 v3LinearVelocity = object.GetLinearVelocity();
		object.SetLinearVelocity(v3AverageLinearVelocity * 0.001f + v3LinearVelocity * 0.999f);
	}	
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
const Vec3& RagDoll::GetPosition()
{
	return m_ppObjectList[SEGMENT_BODY]->GetPosition();
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::ApplyTestImpulse(const Vec3& v3WorldImpulse, const Vec3& v3WorldPosition)
{
	Physics::GetInstance().SetDynamicObjectIsMoving(m_ppObjectList[SEGMENT_BODY]);
	m_ppObjectList[SEGMENT_BODY]->ApplyImpulse(v3WorldImpulse, v3WorldPosition);
	m_ppObjectList[SEGMENT_BODY_LOWER]->ApplyImpulse(v3WorldImpulse, v3WorldPosition);
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::ApplyTestAngularImpulse(const Vec3& v3WorldAngularImpulse)
{
	Physics::GetInstance().SetDynamicObjectIsMoving(m_ppObjectList[SEGMENT_BODY]);
	m_ppObjectList[SEGMENT_BODY]->ApplyAngularImpulse(v3WorldAngularImpulse);
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::SetFrame(const MFrame& frame)
{
	for (int i = 0; i < NUM_SEGMENTS; i++)
	{
		m_ppObjectList[i]->Clear();
		m_ppObjectList[i]->SetFrame(m_pFrameList[i] * frame);
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::AddToPhysics()
{
	for (int i = 0; i < NUM_SEGMENTS; i++)
	{
		Physics::GetInstance().AddDynamicObject(m_ppObjectList[i]);
		m_ppObjectList[i]->SetToMoving(DynamicObject::SET_MOVING_PRIORITY_HIGH);
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::RemoveFromPhysics()
{
	for (int i = 0; i < NUM_SEGMENTS; i++)
	{
		Physics::GetInstance().RemoveDynamicObject(m_ppObjectList[i]);
	}
}

//---------------------------------------------------------------------------------
//---------------------------------------------------------------------------------
void RagDoll::DestroyDynamicObject(DynamicObject*& pDynamicObject)
{
	if (pDynamicObject)
	{
		Physics::GetInstance().RemoveDynamicObject(pDynamicObject);
		pDynamicObject->Release();
		pDynamicObject = 0;	
	}
}
