//---------------------------------------------------------------------------------
// File Name: RagDoll.h
// Description:
//
// Copyright (C) 2004 - 2006 True Axis Pty Ltd, Australia. 
// All Rights Reserved.
//
// Use of this code is covered by the "TRUE AXIS PHYSICS SDK DEMOS SOURCE LICENSE AGREEMENT".
// Details of this agreement are found in the file "TrueAxisPhysicsSDKDemosSourceLicence.txt".
// You have permission to modify this file and use it for any purpose under the terms of
// the agreement.
//
// History:
//		Created File.
//---------------------------------------------------------------------------------

#ifndef TA_RAGDOLL_H
#define TA_RAGDOLL_H

#ifndef TA_MFRAME_H
#include "../TA/Common/MFrame.h"
#endif // TA_MFRAME_H

namespace TA { class DynamicObject; }

class RagDoll
{
public:
	RagDoll();
	~RagDoll();

	void Initialise();
	void Finalise();

	void Render();
	void Update(float fDt);

	const TA::Vec3& GetPosition();
	void ApplyTestImpulse(const TA::Vec3& v3WorldImpulse, const TA::Vec3& v3WorldPosition);
	void ApplyTestAngularImpulse(const TA::Vec3& v3WorldAngularImpulse);
	void SetFrame(const TA::MFrame& frame);
	void AddToPhysics();
	void RemoveFromPhysics();

private:
	enum Segment
	{
		SEGMENT_HEAD = 0,
		SEGMENT_BODY,
		SEGMENT_BODY_LOWER,
		SEGMENT_ARM_UPPER_LEFT,
		SEGMENT_ARM_UPPER_RIGHT,
		SEGMENT_ARM_LOWER_LEFT,
		SEGMENT_ARM_LOWER_RIGHT,
		SEGMENT_LEG_UPPER_LEFT,
		SEGMENT_LEG_UPPER_RIGHT,
		SEGMENT_LEG_LOWER_LEFT,
		SEGMENT_LEG_LOWER_RIGHT,
		NUM_SEGMENTS,
	};

	TA::DynamicObject* m_ppObjectList[NUM_SEGMENTS];
	TA::MFrame m_pFrameList[NUM_SEGMENTS];

	void DestroyDynamicObject(TA::DynamicObject*& pDynamicObject);
};

#endif // TA_RAGDOLL_H
